import simplejson as json


class PlotResult:
    """A class that represents the plot result"""

    def __init__(self, plot):
        self.plot = plot

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, ignore_nan=True)



import sys
import traceback
import threading
import signal
import traceback
import sentry_sdk
from libs import logs
import pandas as pd
from libs import logs
import simplejson as json
from libs.config import read_config
from code_editor import code_editor as ce
from dataframe_manager import dataframe_manager as dm
from experiment_manager import experiment_manager as em
from model_manager import model_manager as mm
from cassist import cassist as ca
from file_explorer import file_explorer as fe
from file_manager import file_manager as fm
from libs.zmq_message import MessageQueuePush, MessageQueuePull
from libs.message import Message, WebappEndpoint, KernelManagerCommand, ExecutorType
from libs.message_handler import BaseMessageHandler
from libs.constants import TrackingModelType
from user_space.user_space import IPythonUserSpace, BaseKernelUserSpace
import cnextlib.dataframe as cd


log = logs.get_logger(__name__)


def kernel_control_handler(user_space):
    log.info("Kernel control thread started")
    ## reading config here to get the most updated version #
    config = read_config('.server.yaml')
    n2p_queue = MessageQueuePull(
        config.n2p_comm['host'], config.n2p_comm['kernel_control_port'])
    try:
        while True:
            strMessage = n2p_queue.receive_msg()
            message = Message(**json.loads(strMessage))
            log.info("Received control message: %s" % message)
            if message.command_name == KernelManagerCommand.restart_kernel:
                result = user_space.restart_executor()
                if result:
                    # get the lastest config to make sure that it is updated with the lastest open project
                    config = read_config('.server.yaml')
                    set_executor_working_dir(user_space, config)
            elif message.command_name == KernelManagerCommand.interrupt_kernel:
                result = user_space.interrupt_executor()
    except:
        trace = traceback.format_exc()
        log.info("Exception %s" % (trace))
        exit(1)


class ShutdownSignalHandler:
    running = True

    def __init__(self, message_handler, user_space):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)
        self.user_space = user_space
        self.message_handler = message_handler

    def get_running_status(self):
        return self.running

    def exit_gracefully(self, *args):
        log.info('ShutdownSignalHandler {}'.format(args))

        if self.message_handler != None:
            for key, value in self.message_handler.items():
                log.info('Shutdown {}'.format(key))
                value.shutdown()
                # value.user_space.executor.interupt_kernel()

        self.running = False
        # currently we exit right here. In the future, consider option to stop message handler gracefully.
        sys.exit(0)


def get_active_project_path(config):
    project_path = None
    for project in config.projects['open_projects']:
        if config.projects['active_project'] == project['id']:
            project_path = project['path']
    return project_path


def set_executor_working_dir(user_space, config):
    project_path = get_active_project_path(config)
    log.info('Project path: {}'.format(project_path))
    if project_path:
        user_space.set_executor_working_dir(project_path)


def main(argv):
    try:
        if argv and len(argv) > 0:
            config = read_config('.server.yaml')

            executor_type = argv[0]
            user_space = None
            message_handler = None
            try:
                p2n_queue = MessageQueuePush(
                    config.p2n_comm['host'], config.p2n_comm['port'])
                if executor_type == ExecutorType.CODE:
                    user_space = IPythonUserSpace(
                        (cd.DataFrame, pd.DataFrame), (TrackingModelType.PYTORCH_NN, TrackingModelType.TENSORFLOW_KERAS))

                    # Start kernel control thread
                    kernel_control_thread = threading.Thread(
                        target=kernel_control_handler, args=(user_space,), daemon=True)
                    kernel_control_thread.start()

                    message_handler = {
                        WebappEndpoint.CodeEditor: ce.MessageHandler(p2n_queue, user_space),
                        WebappEndpoint.DFManager: dm.MessageHandler(p2n_queue, user_space),
                        WebappEndpoint.ModelManager: mm.MessageHandler(p2n_queue, user_space),
                        WebappEndpoint.MagicCommandGen: ca.MessageHandler(
                            p2n_queue, user_space)
                    }

                    set_executor_working_dir(user_space, config)

                elif executor_type == ExecutorType.NONCODE:
                    user_space = BaseKernelUserSpace()
                    message_handler = {
                        WebappEndpoint.ExperimentManager: em.MessageHandler(p2n_queue, user_space),
                        WebappEndpoint.FileManager: fm.MessageHandler(p2n_queue, user_space, config),
                        WebappEndpoint.FileExplorer: fe.MessageHandler(
                            p2n_queue, user_space)
                    }

            except Exception as error:
                log.error("%s - %s" % (error, traceback.format_exc()))
                exit(0)

            shutdowHandler = ShutdownSignalHandler(message_handler, user_space)
            # this condition here is meaningless for now because the process will be exit inside ShutdownSignalHandler.exit_gracefully already
            try:
                while shutdowHandler.running:
                    for line in sys.stdin:
                        try:
                            # log.info('Got message %s' % line)
                            message = Message(**json.loads(line))
                            log.info('Got message from %s command %s' %
                                     (message.webapp_endpoint, message.command_name))
                            if message.webapp_endpoint == WebappEndpoint.KernelManager:
                                if message.command_name == KernelManagerCommand.interrupt_kernel:
                                    shutdowHandler.exit_gracefully()
                            else:
                                message_handler[message.webapp_endpoint].handle_message(
                                    message)
                        except OSError as error:  # TODO check if this has to do with buffer error
                            # since this error might be related to the pipe, we do not send this error to nodejs
                            log.error("OSError: %s" % (error))

                        except:
                            log.error("Failed to execute the command %s",
                                      traceback.format_exc())
                            message = BaseMessageHandler._create_error_message(
                                message.webapp_endpoint, traceback.format_exc(), message.command_name)
                            # send_to_node(message)
                            BaseMessageHandler.send_message(
                                p2n_queue, message.toJSON())

                        try:
                            sys.stdout.flush()
                        except Exception as error:
                            log.error("Failed to flush stdout %s - %s" %
                                      (error, traceback.format_exc()))
            except Exception as error:
                log.error("Exception %s - %s" %
                          (error, traceback.format_exc()))
                ## make sure all the message handler shut down properly #
                shutdowHandler.exit_gracefully()
                exit(0)

    except Exception as error:
        log.error("Failed to flush stdout %s - %s" %
                  (error, traceback.format_exc()))
        exit(0)


if __name__ == "__main__":

    # Initialize sentry
    sentry_sdk.init(
        "https://318e8d1b2b0b43fbbbd38ea0426378b4@o1259763.ingest.sentry.io/6435268",

        # Set traces_sample_rate to 1.0 to capture 100%
        # of transactions for performance monitoring.
        # We recommend adjusting this value in production.
        traces_sample_rate=0.1
    )

    main(sys.argv[1:])



import os
import simplejson as json


class DirMetatdata:
    # def __init__(self, **entries):
    #     self.path = None
    #     self.name = None
    #     self.is_file = None
    #     self.__dict__.update(entries)

    def __init__(self, path, name, is_file, deletable, timestamp):
        self.path = path
        self.name = name
        self.is_file = is_file
        self.deletable = deletable
        self.timestamp = timestamp

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, ignore_nan=True)

    def __repr__(self) -> str:
        return self.toJSON()


class FileMetadata:
    # def __init__(self, **entries):
    #     self.path = None
    #     self.name = None
    #     # self.type = None
    #     self.executor = None
    #     self.timestamp = None
    #     self.__dict__.update(entries)

    def __init__(self, path, name, timestamp=None, executor=False):
        self.path = path
        self.name = name
        # self.type = None
        self.executor = executor
        if timestamp != None:
            self.timestamp = timestamp

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, ignore_nan=True)

    def __repr__(self) -> str:
        return self.toJSON()


class FileContent:
    def __init__(self, content, code_lines=None, timestamp=None):
        # self.__dict__.update(locals())
        self.content = content
        self.code_lines = code_lines
        self.timestamp = timestamp

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, ignore_nan=True)

    def __repr__(self) -> str:
        return self.toJSON()


class ProjectMetadata:
    def __init__(self, **entries):
        self.path = None
        self.name = None
        self.id = None
        self.__dict__.update(entries)
        self.data_path = os.path.join(self.path, '.cnext')
        self.config_path = os.path.join(self.data_path, 'cnext.yaml')

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, ignore_nan=True)

    def __repr__(self) -> str:
        return self.toJSON()



import os
from os.path import exists
import simplejson as json
import traceback
from project_manager.interfaces import ProjectMetadata
from libs import logs
from libs.config import read_config, save_config
from project_manager.interfaces import FileMetadata

log = logs.get_logger(__name__)

active_project = None
CNEXT_PROJECT_DIR = ''
FILE_CONFIG = 'config.json'


def get_open_files():
    open_files = []
    try:
        if active_project:
            config_path = active_project.config_path  # active_project.path+'/.cnext.yaml'
            if exists(config_path):
                config = read_config(config_path)
                if hasattr(config, 'open_files'):
                    open_files = config.open_files
            else:
                log.error("Config file does not exist %s" % (config_path))
        return open_files
    # except yaml.YAMLError as error:
    #     log.error("%s" % (error))
    #     return []
    except Exception as error:
        log.error("%s - %s" % (error, traceback.format_exc()))
        raise Exception  # this will be seen in the web app #


def close_file(path):
    open_files = []
    try:
        if active_project:
            config_path = active_project.config_path  # active_project.path+'/.cnext.yaml'
            if exists(config_path):
                config = read_config(config_path)
                if hasattr(config, 'open_files'):
                    for f in config.open_files:
                        ## do not include the file that is being closed
                        if(f['path'] != path):
                            open_files.append(f)
                config.open_files = open_files
                save_config(config.__dict__, config_path)
            else:
                log.error("Config file does not exist %s" % (config_path))
        return open_files
    # except yaml.YAMLError as error:
    #     log.error("%s" % (error))
    #     return []
    except Exception as error:
        log.error("%s - %s" % (error, traceback.format_exc()))
        raise Exception  # this will be seen in the web app #


def open_file(path):
    open_files = []
    try:
        if active_project:
            config_path = active_project.config_path  # active_project.path+'/.cnext.yaml'
            if exists(config_path):
                config = read_config(config_path)
                if hasattr(config, 'open_files'):
                    open_files = config.open_files
                ## Note that we dont set the timestamp when open the file #
                file = FileMetadata(path,
                                    name=path.split('/')[-1],
                                    executor=(config.executor == path))
                open_files.append(file.__dict__)
                config.open_files = open_files
                save_config(config.__dict__, config_path)
            else:
                log.error("Config file does not exist %s" % (config_path))
        return open_files
    # except yaml.YAMLError as error:
    #     log.error("%s" % (error))
    #     return []
    except Exception as error:
        log.error("%s - %s" % (error, traceback.format_exc()))
        raise Exception  # this will be seen in the web app #


def set_project_dir(path):
    global CNEXT_PROJECT_DIR
    CNEXT_PROJECT_DIR = path
    return True


def set_active_project(project: ProjectMetadata):
    global active_project
    active_project = project
    return True


def get_active_project():
    return active_project


def set_working_dir(path):
    try:
        os.chdir(path)
        return True
    except Exception:
        raise Exception


def save_project_config(content):
    try:
        # config_file_path = os.path.join(project_path, FILE_CONFIG)
        active_project = get_active_project()
        config_file_path = os.path.join(
            active_project.path, FILE_CONFIG)
        os.makedirs(os.path.dirname(config_file_path), exist_ok=True)
        with open(config_file_path, 'w') as outfile:
            outfile.write(json.dumps(content, indent=4))
        return FileMetadata(config_file_path, name=FILE_CONFIG, timestamp=os.path.getmtime(config_file_path))
    except Exception as ex:
        log.error("Save config error: {}".format(ex))
        raise Exception


def get_project_config():
    active_project = get_active_project()
    config_file_path = os.path.join(
        active_project.path, FILE_CONFIG)
    if os.path.exists(config_file_path):
        config_file_data = open(config_file_path, "r")
        data = json.loads(config_file_data.read())
        return data
    return



import threading
import traceback
import simplejson as json

import plotly
from libs.message_handler import BaseMessageHandler
from libs.message import ContentType, SubContentType, Message

from libs import logs
from libs.message import DFManagerCommand, WebappEndpoint, CodeEditorCommand, ModelManagerCommand
from user_space.ipython.constants import IPythonConstants, IpythonResultMessage
log = logs.get_logger(__name__)


class MessageHandler(BaseMessageHandler):
    def __init__(self, p2n_queue, user_space=None):
        super(MessageHandler, self).__init__(p2n_queue, user_space)

    # @staticmethod
    # def _is_execute_result(header) -> bool:
    #     return header['msg_type'] == IPythonConstants.MessageType.EXECUTE_RESULT

    # @staticmethod
    # def _is_execute_reply(header) -> bool:
    #     return header['msg_type'] == IPythonConstants.MessageType.EXECUTE_REPLY

    # @staticmethod
    # def _is_stream_result(header) -> bool:
    #     return header['msg_type'] == IPythonConstants.MessageType.STREAM

    # @staticmethod
    # def _is_display_data_result(header) -> bool:
    #     return header['msg_type'] == IPythonConstants.MessageType.DISPLAY_DATA

    # @staticmethod
    # def _is_error_message(header) -> bool:
    #     return header['msg_type'] == IPythonConstants.MessageType.ERROR

    @staticmethod
    def _result_is_plotly_fig(content) -> bool:
        try:
            # Because IPython return plotly as json format, have to convert it to figure instance
            # Use plotly_io (low-level interface for displaying, reading and writing figures)
            plotly_figure = plotly.io.from_json(json.dumps(content))
            return hasattr(plotly.graph_objs, '_figure') and (type(plotly_figure) == plotly.graph_objs._figure.Figure)
        except Exception:
            return False

    # PRIORITIXED_MIME_LIST = [SubContentType.APPLICATION_PLOTLY, SubContentType.APPLICATION_JSON, SubContentType.IMAGE_PLOTLY,
    #                          SubContentType.IMAGE_JPG, SubContentType.IMAGE_PNG, SubContentType.IMAGE_SVG, SubContentType.TEXT_HTML]

    def _process_rich_ouput(self, message, msg_ipython):
        message.error = False
        if type(msg_ipython.content['data']) is dict:
            message.type = ContentType.RICH_OUTPUT
        else:
            message.type = ContentType.STRING
        message.content = msg_ipython.content['data']

        # remove 'text/html' key if the output is plotly to improve the efficiency.
        # TODO: revisit this later #
        if type(message.content) is dict and SubContentType.APPLICATION_PLOTLY in message.content:
            message.content.pop('text/html', None)
        return message

    def _process_error_message(self, message, ipython_msg):
        # log.error("Error %s" % (msg_ipython.content['traceback']))
        if isinstance(ipython_msg.content['traceback'], list):
            content = '\n'.join(ipython_msg.content['traceback'])
        else:
            content = ipython_msg.content['traceback']
        return self._create_error_message(
            WebappEndpoint.CodeEditor, content, message.command_name, message.metadata)

    def _process_stream_message(self, message, ipython_msg):
        message.error = False
        message.type = ContentType.STRING
        if 'text' in ipython_msg.content:
            message.content = ipython_msg.content['text']
        elif 'data' in ipython_msg.content:
            message.content = ipython_msg.content['data']
        return message

    def _process_other_message(self, message, ipython_msg):
        message.error = False
        message.type = ContentType.IPYTHON_MSG
        message.sub_type = SubContentType.NONE
        message.content = ipython_msg.content
        return message

    def _create_return_message(self, ipython_message, stream_type, client_message):
        """
            Get single message from IPython,
            classify it according to the message type then return it to the client
        """
        ipython_message = IpythonResultMessage(**ipython_message)
        message = Message(**{'webapp_endpoint': WebappEndpoint.CodeEditor, 'command_name': client_message.command_name})

        log.info('Got message from ipython: %s %s',
                 ipython_message.header['msg_type'], ipython_message.content['status'] if 'status' in ipython_message.content else None)

        # Add header message from ipython to message metadata
        if message.metadata == None:
            message.metadata = {}

        if client_message.metadata != None:
            message.metadata.update(client_message.metadata)
        message.metadata.update(dict((k, ipython_message.header[k])
                                     for k in ('msg_id', 'msg_type', 'session')))
        message.metadata.update({'stream_type': stream_type})

        if self._is_error_message(ipython_message.header):
            message = self._process_error_message(message, ipython_message)
        elif self._is_stream_result(ipython_message.header):
            message = self._process_stream_message(message, ipython_message)
        elif self._is_execute_result(ipython_message.header) or self._is_display_data_result(ipython_message.header):
            message = self._process_rich_ouput(message, ipython_message)
        else:
            message = self._process_other_message(message, ipython_message)

        return message

    def message_handler_callback(self, ipython_message, stream_type, client_message):
        try:
            # if self.request_metadata is not None:
            message = self._create_return_message(
                ipython_message=ipython_message, stream_type=stream_type, client_message=client_message)            
            # log.info('Reply message: %s' % message)
            if message != None:
                self._send_to_node(message)
        except:
            trace = traceback.format_exc()
            log.info("Exception %s" % (trace))
            error_message = BaseMessageHandler._create_error_message(
                client_message.webapp_endpoint, trace, client_message.command_name, {})
            self._send_to_node(error_message)

    def handle_message(self, message):
        try:
            self.user_space.execute(
                message.content, None, self.message_handler_callback, client_message=message)
            self._get_active_dfs_status()
            # self._get_active_model()
        except:
            trace = traceback.format_exc()
            log.info("Exception %s" % (trace))            
            error_message = BaseMessageHandler._create_error_message(message.webapp_endpoint, trace, message.command_name, {})
            self._send_to_node(error_message)

    def _get_active_dfs_status(self):
        active_df_status = self.user_space.get_active_dfs_status()
        active_df_status_message = Message(**{"webapp_endpoint": WebappEndpoint.DFManager, "command_name": DFManagerCommand.update_df_status,
                                              "seq_number": 1, "type": "dict", "content": active_df_status, "error": False})
        self._send_to_node(active_df_status_message)

    def _get_active_models_info(self):
        active_models = self.user_space.get_active_models_info()
        active_models_message = Message(**{"webapp_endpoint": WebappEndpoint.ModelManager, "command_name": ModelManagerCommand.get_active_models_info,
                                              "seq_number": 1, "type": "dict", "content": active_models, "error": False})
        self._send_to_node(active_models_message)



import io
import base64
import traceback
import simplejson as json
from xmlrpc.client import boolean

import pandas
import plotly
import matplotlib.pyplot as plt
from libs.message_handler import BaseMessageHandler
from libs.message import ContentType, Message

from libs import logs
from libs.message import DFManagerCommand
from libs.json_serializable import ipython_internal_output
from libs.message import SubContentType
from user_space.ipython.constants import IPythonInteral, IpythonResultMessage
from user_space.user_space import ExecutionMode

log = logs.get_logger(__name__)


class MessageHandler(BaseMessageHandler):
    def __init__(self, p2n_queue, user_space=None):
        super(MessageHandler, self).__init__(p2n_queue, user_space)

    def _create_get_cardinal_data(self, result):
        return {'cardinals': result}

    @ipython_internal_output
    def _get_cardinal(self, metadata):
        df_id = metadata['df_id']
        col_name = metadata['col_name']
        if 'groupby' in metadata:
            groupby = metadata['groupby']

            # Get stat for groupby x0
            groupby_cols = ''
            groupby_cols += '"%s",' % groupby[0]
            command_str = '%s.groupby([%s])["%s"].count()' % (
                df_id, groupby_cols, col_name)
            # result = {'groupby_x0': eval(
            #     command_str, globals()).describe().to_dict()}
            result = {'groupby_x0': self.user_space.execute(
                command_str, ExecutionMode.EVAL).describe().to_dict()}
            # Get stat for groupby all x
            groupby_cols = ''
            for c in groupby:
                groupby_cols += '"%s",' % c
            command_str = '%s.groupby([%s])["%s"].count()' % (
                df_id, groupby_cols, col_name)
            # result.update({'groupby_all': eval(
            #     command_str, globals()).describe().to_dict()})
            result = {'groupby_all': self.user_space.execute(
                command_str, ExecutionMode.EVAL).describe().to_dict()}
            # Count unique and get monotonic for all x
            unique_counts = {}
            monotonics = {}
            for col_name in groupby:
                command_str = 'len(%s["%s"].unique())' % (
                    df_id, col_name)
                # unique_counts[col_name] = eval(command_str, globals())
                unique_counts[col_name] = self.user_space.execute(
                    command_str, ExecutionMode.EVAL)
                command_str = '%s["%s"].is_monotonic' % (
                    df_id, col_name)
                # monotonics[col_name] = eval(command_str, globals())
                # monotonics[col_name] = eval(command_str, globals())
                monotonics[col_name] = self.user_space.execute(
                    command_str, ExecutionMode.EVAL)
            result.update({'unique_counts': unique_counts,
                           'monotonics': monotonics})
            # log.info('Result:  %s' % (result))
        else:
            # TODO: consider to remove this because it is not used
            # return a describe dict to make it consistent with the groupby case above
            command_str = '%s["%s"].shape[0]' % (df_id, col_name)
            # result = pandas.DataFrame(
            #     [eval(command_str, globals())]).describe().to_dict()
            result = pandas.DataFrame(
                [self.user_space.execute(command_str, ExecutionMode.EVAL)]).describe().to_dict()

        if result:
            return self._create_get_cardinal_data(result)
        else:
            return None

    def handle_message(self, message):
        send_reply = False
        # message execution_mode will always be `eval` for this sender
        try:
            if message.command_name == DFManagerCommand.get_cardinal:
                # message.metadata = message.metadata.replace("'", '"')
                output_messages = self.user_space.execute(
                    "{}._get_cardinal({})".format(IPythonInteral.CASSIST.value, message.metadata))
                ipython_message = IpythonResultMessage(**output_messages)
                output = self.get_execute_result(ipython_message)
                if output is not None:
                    # log.info("get table data")
                    log.info('cAssist get cardinal: %s' % output)
                    type = ContentType.COLUMN_CARDINAL
                    sub_type = SubContentType.NONE
                    send_reply = True

            if send_reply:
                message.type = type
                message.content = output
                message.error = False
                self._send_to_node(message)
        except:
            trace = traceback.format_exc()
            log.error("%s" % (trace))
            error_message = MessageHandler._create_error_message(
                message.webapp_endpoint, trace)
            self._send_to_node(error_message)



import importlib
import os
import tempfile
import threading
import time
import traceback
import socket
import netron
import requests

from libs.message_handler import BaseMessageHandler
from libs import logs
from libs.message import Message, WebappEndpoint, ModelManagerCommand
from model_manager.interfaces import ModelInfo, NetronStatus
from user_space.ipython.constants import IPythonConstants
log = logs.get_logger(__name__)


class MessageHandler(BaseMessageHandler):
    def __init__(self, p2n_queue,  user_space=None):
        super(MessageHandler, self).__init__(p2n_queue, user_space)
        self.execution_lock = threading.Lock()
        sock = socket.socket()
        sock.bind(('', 0))
        self.netron_port = sock.getsockname()[1]
        self.netron_host = 'localhost'
        self.netron_tmp_dir = tempfile.gettempdir()
        self.netron_address = None

    @classmethod
    def _fullname(cls, klass):
        module = klass.__module__
        if module == 'builtins':
            return klass.__qualname__
        return module + '.' + klass.__qualname__

    def _complete_execution_message(self, message) -> bool:
        return message['header']['msg_type'] == 'execute_reply' and 'status' in message['content']

    def _message_handler_callback(self, ipython_message, stream_type, client_message):
        try:
            log.info('%s msg: %s %s' % (
                stream_type, ipython_message['header']['msg_type'], ipython_message['content']))
            if self._complete_execution_message(ipython_message) and self.execution_lock.locked():
                self.execution_status = ipython_message['content']['status']
                self.execution_lock.release()
                log.info('Execution unlocked')
            else:
                # TODO: log everything else
                log.info('Other messages: %s' % ipython_message)
        except:
            # this is internal exception, we won't send it to the client
            trace = traceback.format_exc()
            log.info("Exception %s" % (trace))

    def _load_class(name):
        components = name.split('.')
        mod = importlib.import_module(components[0])
        for comp in components[1:]:
            mod = getattr(mod, comp)
        return mod

    def _complete_waiting_execution(func):
        '''
        Wrapper to block the execution until the execution complete
        '''
        def _complete_waiting_execution_wrapper(*args, **kwargs):
            args[0].execution_lock.acquire()
            log.info('Execution locked')
            result = func(*args, **kwargs)
            args[0].execution_lock.acquire()
            args[0].execution_lock.release()
            return {'status': args[0].execution_status, 'result': result}
        return _complete_waiting_execution_wrapper

    @_complete_waiting_execution
    def _save_model(self, modelInfo: ModelInfo):
        MODEL_PATH = None
        if modelInfo.base_class == "keras.engine.training.Model":
            MODEL_PATH = os.path.join(
                self.netron_tmp_dir, '{}.h5'.format(modelInfo.name))
            if os.path.exists(MODEL_PATH):
                os.remove(MODEL_PATH)
            code = '{}.save("{}")'.format(modelInfo.name, MODEL_PATH)
            self.user_space.execute(code, None, self._message_handler_callback)
        elif modelInfo.base_class == "torch.nn.modules.module.Module":
            MODEL_PATH = os.path.join(
                self.netron_tmp_dir, '{}.onnx'.format(modelInfo.name))
            if os.path.exists(MODEL_PATH):
                os.remove(MODEL_PATH)
            code = 'torch.onnx.export({}, {}.createInput(), f="{}", training=False)'.format(
                modelInfo.name, modelInfo.name, MODEL_PATH)
            self.user_space.execute(code, None, self._message_handler_callback)
        return MODEL_PATH

    def _start_netron_server(self, model_path) -> NetronStatus:
        self.netron_address = (self.netron_host, self.netron_port)
        netron.stop(self.netron_address)
        netron.start(model_path, self.netron_address, browse=False)
        # wait until the server is available or timeout
        TIMEOUT = 5000
        SLEEP_TIME = 50
        time_pass = 0
        while (time_pass := time_pass+SLEEP_TIME) < TIMEOUT:
            response = requests.get(
                'http://%s:%d' % (self.netron_address[0], self.netron_address[1]))
            if response.status_code == 200:
                return NetronStatus.OK
            else:
                time.sleep(SLEEP_TIME);
        return NetronStatus.ERROR 

    def _display_model(self, modelInfo: ModelInfo):
        # torch.nn.Module, tensorflow.keras.Model
        output = self._save_model(modelInfo)
        # log.info("Code to run: {}".format(code))
        if output['status'] == IPythonConstants.IOBufMessageStatus.OK:
            netron_status = self._start_netron_server(output['result'])
            return {'address': self.netron_address, 'status': netron_status}

    def handle_message(self, message):
        # send_reply = False
        # message execution_mode will always be `eval` for this sender
        log.info('Got message %s' % message)
        # log.info('Globals: %s' % client_globals)

        ## this step is important to make sure the query work properly in the backend #
        if (isinstance(message.content, str)):
            message.content = message.content.replace("'", '"')

        try:
            if message.command_name == ModelManagerCommand.get_active_models_info:
                active_models_info = self.user_space.get_active_models_info()
                active_models_info_message = Message(**{"webapp_endpoint": WebappEndpoint.ModelManager, "command_name": message.command_name,
                                                        "seq_number": 1, "type": "dict", "content": active_models_info, "error": False})
                self._send_to_node(active_models_info_message)
            if message.command_name == ModelManagerCommand.display_model:
                modelInfo = ModelInfo(**message.content)
                display_info = self._display_model(modelInfo)
                display_info_message = Message(**{"webapp_endpoint": WebappEndpoint.ModelManager, "command_name": message.command_name,
                                                  "seq_number": 1, "type": "dict", "content": display_info, "error": False})
                self._send_to_node(display_info_message)

        except:
            trace = traceback.format_exc()
            log.error("%s" % (trace))
            error_message = MessageHandler._create_error_message(
                message.webapp_endpoint, trace, message.command_name)
            self._send_to_node(error_message)

    def shutdown(self):
        log.info('Netron stop at address {}'.format(self.netron_address))
        if netron.status(self.netron_address):
            log.info('Netron stop at address {}'.format(self.netron_address))
            netron.stop(self.netron_address)
        return super().shutdown()



import simplejson as json
from enum import Enum

# logging.basicConfig(filename='./log.txt', filemode='a', format='%(asctime)s,%(msecs)d %(name)s %(funcName)s %(levelname)s %(message)s',
#                         datefmt='%H:%M:%S', level=logging.DEBUG)
# log = logging.getLogger(__name__)


class WebappEndpoint(str, Enum):
    DFManager = 'DFManager'
    CodeEditor = 'CodeEditor'
    FileManager = 'FileManager'
    MagicCommandGen = 'MagicCommandGen'
    FileExplorer = 'FileExplorer'
    ExperimentManager = 'ExperimentManager'
    KernelManager = 'KernelManager'
    ModelManager = 'ModelManager'

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)


class DFManagerCommand(str, Enum):
    update_df_status = 'update_df_status'
    reload_df_status = 'reload_df_status'
    plot_column_histogram = 'plot_column_histogram'
    get_countna = 'get_countna'
    plot_countna = 'plot_countna'
    get_table_data = 'get_table_data'
    get_df_metadata = 'get_df_metadata'
    plot_column_quantile = 'plot_column_quantile'
    get_cardinal = 'get_cardinal'
    get_file_content = 'get_file_content'

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)


class ProjectCommand(str, Enum):
    list_dir = 'list_dir'
    get_file_metadata = 'get_file_metadata'
    read_file = 'read_file'
    save_file = 'save_file'
    close_file = 'close_file'
    open_file = 'open_file'
    set_name = 'set_name'
    get_open_files = 'get_open_files'
    create_file = 'create_file'
    create_folder = 'create_folder'
    delete = 'delete'
    remove_file = 'remove_file'
    remove_folder = 'remove_folder'
    set_working_dir = 'set_working_dir'
    set_project_dir = 'set_project_dir'
    get_active_project = 'get_active_project'
    save_state = 'save_state'
    save_project_config = 'save_project_config'
    get_project_config = 'get_project_config'


class ExperimentManagerCommand(str, Enum):
    list_experiments = 'list_experiments'
    list_run_infos = 'list_run_infos'
    get_metric_plots = 'get_metric_plots'
    load_artifacts_to_local = "load_artifacts_to_local"
    set_tracking_uri = 'set_tracking_uri'
    ## get data for a list of runs #
    get_runs_data = 'get_runs_data'


class KernelManagerCommand(str, Enum):
    restart_kernel = 'restart_kernel'
    interrupt_kernel = 'interrupt_kernel'


class ModelManagerCommand(str, Enum):
    get_active_models_info = 'get_active_models_info'
    display_model = 'display_model'


class CodeEditorCommand(str, Enum):
    exec_line = 'exec_line'
    exec_grouped_lines = 'exec_grouped_lines'


class ContentType(str, Enum):
    COMMAND = 'command'
    STRING = 'str'
    DICT = 'dict'
    BINARY = 'binary'
    PANDAS_DATAFRAME = 'pandas_dataframe'
    DIR_LIST = 'dir_list'
    FILE_METADATA = 'file_metadata'
    PROJECT_METADATA = 'project_metadata'
    FILE_CONTENT = 'file_content'
    COLUMN_CARDINAL = 'column_cardinal'
    RICH_OUTPUT = 'rich_output'
    PLOTLY_FIG = 'plotly_fig'  # It is used for code editor basekernel
    IPYTHON_MSG = 'ipython_msg'
    NONE = 'none'

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)


class SubContentType(str, Enum):
    IMAGE_PLOTLY = 'image/plotly+json'
    IMAGE_SVG = 'image/svg+xml'
    IMAGE_PNG = 'image/png'
    IMAGE_JPG = 'image/jpg'
    TEXT_HTML = 'text/html'
    APPLICATION_JSON = 'application/json'
    APPLICATION_CNEXT = 'application/cnext+json'
    APPLICATION_PLOTLY = 'application/vnd.plotly.v1+json'

    NONE = 'none'

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)


class CommandType(str, Enum):
    MFLOW = 'mlflow'  # use mlflow object to call the function #
    MLFLOW_CLIENT = 'mlflow_client'  # use mlflow.client object to call the function #
    MLFLOW_OTHERS = 'mlflow_others'  # commands that requires more than standard mlflow api #

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)


class Message:
    def __init__(self, **entries):
        self.webapp_endpoint = None
        self.command_name = None
        self.seq_number = None
        self.type = None
        self.sub_type = None
        self.content = None
        self.error = None
        self.metadata = None
        self.__dict__.update(entries)

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, ignore_nan=True)

    def __repr__(self) -> str:
        return self.toJSON()


class ExecutorType(str, Enum):
    CODE = 'code'
    NONCODE = 'noncode'



import io
import base64
import traceback
import plotly
import matplotlib.pyplot as plt
from cnextlib.df_status_hook import DataFrameStatusHook
import cnextlib.dataframe as cd
from libs.message_handler import BaseMessageHandler
from libs.message import ContentType, Message, SubContentType

from libs import logs
from libs.message import DFManagerCommand, WebappEndpoint
from user_space.user_space import ExecutionMode
from user_space.user_space import BaseKernel, IPythonUserSpace
log = logs.get_logger(__name__)


class MessageHandler(BaseMessageHandler):
    def __init__(self, p2n_queue, user_space=None):
        super(MessageHandler, self).__init__(p2n_queue, user_space)

    @staticmethod
    def _create_plot_data(result):
        return result.to_json()

    @staticmethod
    def _create_matplotlib_data(result):
        figfile = io.BytesIO()
        plt.savefig(figfile, format='svg')
        figfile.seek(0)  # rewind to beginning of file
        plot_binary_data = base64.b64encode(figfile.getvalue())
        return plot_binary_data

    @staticmethod
    def _result_is_plotly_fig(result) -> bool:
        return hasattr(plotly.graph_objs, '_figure') and (type(result) == plotly.graph_objs._figure.Figure)

    @staticmethod
    def _assign_exec_mode(message: Message):
        message.execution_mode = 'eval'
        if message.metadata and ('line_range' in message.metadata):
            line_range = message.metadata['line_range']
            # always 'exec' if there are more than 1 line in the code
            if line_range['fromLine'] < line_range['toLine']-1:
                message.execution_mode = 'exec'

        try:
            compile(message.content, '<stdin>', 'eval')
        except SyntaxError as error:
            log.error(error)
            message.execution_mode = 'exec'

        log.info("assigned command type: %s" % message.execution_mode)

    @staticmethod
    def _result_is_matplotlib_fig(result) -> bool:
        # because matplotlib return array of objects
        if isinstance(result, list) and len(result) > 0:
            matplotlib_object = result[0]
            result_str_type = str(type(matplotlib_object))
            if 'matplotlib' in result_str_type and hasattr(matplotlib_object, 'figure') and matplotlib_object.figure.number > 0:
                return True
        return False

    def _get_active_dfs_status(self):
        active_dfs_status = self.user_space.get_active_dfs_status()
        if active_dfs_status:
            active_df_status_message = Message(**{"webapp_endpoint": WebappEndpoint.DFManager,
                                                  "command_name": DFManagerCommand.update_df_status,
                                                  "seq_number": 1,
                                                  "type": "dict",
                                                  "content": active_dfs_status,
                                                  "error": False})
            self._send_to_node(active_df_status_message)

    def handle_message(self, message):
        # message execution_mode will always be `eval` for this sender
        log.info('eval... %s' % message)
        # log.info('Globals: %s' % client_globals)
        try:
            sub_type = SubContentType.NONE
            self._assign_exec_mode(message)
            type = ContentType.NONE
            output = ''
            if message.execution_mode == 'exec':
                log.info('exec mode...')
                # exec(message.content, globals())
                self.user_space.execute(message.content, ExecutionMode.EXEC)
                type = ContentType.STRING
                # output = sys.stdout.getvalue()
            elif message.execution_mode == 'eval':
                log.info('eval mode...')
                # result = eval(message.content, globals())
                result = self.user_space.execute(
                    message.content, ExecutionMode.EVAL)
                if result is not None:
                    # log.info("eval result: \n%s" % (result))
                    log.info("got eval results")
                    # if self._result_is_dataframe(result):
                    #     df_id = self._get_dataframe_id(message.content)
                    #     output = self._create_table_data(df_id, result)
                    #     type = ContentType.PANDAS_DATAFRAME
                    if self._result_is_plotly_fig(result):
                        output = self._create_plot_data(result)
                        type = ContentType.RICH_OUTPUT
                        sub_type = SubContentType.IMAGE_PLOTLY
                    elif self._result_is_matplotlib_fig(result):
                        output = self._create_matplotlib_data(result)
                        type = ContentType.RICH_OUTPUT
                        sub_type = 'image/svg+xml'
                    else:
                        type = ContentType.STRING
                        output = str(result)
            # log.info('Globals: %s' % globals())

            message.type = type
            message.sub_type = sub_type
            message.content = output
            message.error = False
            self._send_to_node(message)

            self._get_active_dfs_status()

        except:
            trace = traceback.format_exc()
            log.error("Exception %s" % (trace))
            error_message = MessageHandler._create_error_message(
                message.webapp_endpoint, trace, message.command_name, message.metadata)
            self._send_to_node(error_message)



import os
import traceback

from libs.message_handler import BaseMessageHandler
from libs.message import ContentType

from libs import logs
from libs.config import read_config
from libs.message import ProjectCommand
from project_manager import files, projects

log = logs.get_logger(__name__)


class MessageHandler(BaseMessageHandler):
    def __init__(self, p2n_queue, user_space, config):
        super(MessageHandler, self).__init__(p2n_queue, user_space)
        open_projects = []
        active_project: projects.ProjectMetadata = None
        if hasattr(config, 'projects'):
            if 'open_projects' in config.projects:
                if isinstance(config.projects['open_projects'], list):
                    open_projects = config.projects['open_projects']
            if 'active_project' in config.projects:
                for project_config in open_projects:
                    if config.projects['active_project'] == project_config['id']:
                        active_project = projects.ProjectMetadata(
                            **project_config)
        if active_project:
            projects.set_active_project(active_project)

    def handle_message(self, message):
        log.info('FileManager handle message: %s %s %s' %
                 (message.command_name, message.type, message.sub_type))
        try:
            metadata = message.metadata
            if 'path' in metadata.keys():
                ## avoid creating `./` when the path is empty
                if metadata['path'] == "":
                    norm_path = metadata['path']
                else:
                    norm_path = os.path.normpath(metadata['path'])
            if 'project_path' in metadata.keys():
                norm_project_path = os.path.normpath(metadata['project_path'])

            result = None
            if message.command_name == ProjectCommand.list_dir:
                result = []
                if 'path' in metadata.keys() and 'project_path' in metadata.keys():
                    result = files.list_dir(norm_project_path, norm_path)
                    type = ContentType.DIR_LIST
            elif message.command_name == ProjectCommand.get_open_files:
                result = projects.get_open_files()
                type = ContentType.FILE_METADATA
            elif message.command_name == ProjectCommand.set_working_dir:
                if 'path' in metadata.keys():
                    result = projects.set_working_dir(norm_path=os.path.normpath(metadata['path'])
                                                      )
                type = ContentType.NONE
            elif message.command_name == ProjectCommand.set_project_dir:
                if 'path' in metadata.keys():
                    result = projects.set_project_dir(norm_path=os.path.normpath(metadata['path'])
                                                      )
                type = ContentType.NONE
            elif message.command_name == ProjectCommand.read_file:
                if 'path' in metadata.keys() and 'project_path' in metadata.keys():
                    timestamp = metadata['timestamp'] if 'timestamp' in metadata else None
                    result = files.read_file(norm_project_path, norm_path,
                                             timestamp=timestamp)
                    if result == None:
                        type = ContentType.NONE
                    else:
                        type = ContentType.FILE_CONTENT
            elif message.command_name == ProjectCommand.save_file:
                if 'path' in metadata.keys():
                    result = files.save_file(
                        norm_project_path, norm_path, content=message.content)
                type = ContentType.FILE_METADATA
            elif message.command_name == ProjectCommand.close_file:
                if 'path' in metadata.keys():
                    result = projects.close_file(norm_path)
                type = ContentType.FILE_METADATA
            elif message.command_name == ProjectCommand.open_file:
                if 'path' in metadata.keys():
                    result = projects.open_file(norm_path)
                type = ContentType.FILE_METADATA
            elif message.command_name == ProjectCommand.get_active_project:
                result = projects.get_active_project()
                type = ContentType.PROJECT_METADATA
            elif message.command_name == ProjectCommand.save_state:
                if 'path' in metadata.keys() and 'project_path' in metadata.keys():
                    result = files.save_state(
                        norm_project_path, norm_path, content=message.content)
                type = ContentType.FILE_METADATA
            elif message.command_name == ProjectCommand.save_project_config:
                result = projects.save_project_config(content=message.content)
                type = ContentType.PROJECT_METADATA
            elif message.command_name == ProjectCommand.get_project_config:
                result = projects.get_project_config()
                type = ContentType.PROJECT_METADATA

            # create reply message
            message.type = type
            message.content = result
            message.error = False
            self._send_to_node(message)

        except:
            trace = traceback.format_exc()
            log.error("%s" % (trace))
            error_message = MessageHandler._create_error_message(
                message.webapp_endpoint, trace, command_name=message.command_name)
            self._send_to_node(error_message)
